// value & error-bound layer for mpfs
// TODO pow_ui
// FIXME allow f(x,x,x)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpfr.h>
#include "mpfre.h"

static const mp_prec_t error_prec=32;

/* init and clear */

mpfre_t mpfre_init(mp_prec_t p) {
  mpfre_t z=malloc(sizeof(struct _mpfre));
  mpfr_init2(z->x,p);
  mpfr_init2(z->d,error_prec);
  return z;
}

void mpfre_init2(mpfre_t z, mp_prec_t p) {
  mpfr_init2(z->x,p);
  mpfr_init2(z->d,error_prec);
}

void mpfre_clear(mpfre_t x) {
  mpfr_clear(x->x);
  mpfr_clear(x->d);
  free(x);
}

/* set */

void mpfre_set(mpfre_t z, mpfre_t x) {
  mpfr_set(z->x,x->x,GMP_RNDN);
  mpfr_set(z->d,x->d,GMP_RNDU);
}

void mpfre_set_ui(mpfre_t z, unsigned int x, double d) {
  mpfr_set_ui(z->x,x,GMP_RNDN);
  mpfr_set_d(z->d,d,GMP_RNDU);
}

void mpfre_set_si(mpfre_t z, int x, double d) {
  mpfr_set_si(z->x,x,GMP_RNDN);
  mpfr_set_d(z->d,d,GMP_RNDU);
}

void mpfre_set_d(mpfre_t z, double x, double d) {
  mpfr_set_d(z->x,x,GMP_RNDN);
  mpfr_set_d(z->d,d,GMP_RNDU);
}

void mpfre_set_qui(mpfre_t z, unsigned long x, unsigned long y, double d) {
  mpfr_set_ui(z->x,x,GMP_RNDN);
  mpfr_div_ui(z->x,z->x,y,GMP_RNDN);
  mpfr_set_d(z->d,d,GMP_RNDU);
}

void mpfre_set_str(mpfre_t z, char* x, int base, double d) {
  mpfr_set_str(z->x,x,10,GMP_RNDN);
  mpfr_set_d(z->d,d,GMP_RNDU);
}

/* output */

int mpfre_out_str(mpfre_t x) {
  const double log2ten=3.3219280948873623478703194294893901759;
  double v=mpfr_get_d(x->x,GMP_RNDD);
  double d=mpfr_get_d(x->d,GMP_RNDD);
  double p,g=1-log(d/fabs(v))/log(10);
  p=mpfr_get_prec(x->x)/log2ten;
  g=g<p?g:p;
  g=g<2?2:g;
  return
  mpfr_out_str(stdout,10,(mp_prec_t)g,x->x,GMP_RNDN)
  +printf(" +or- ")
  +mpfr_out_str(stdout,10,6,x->d,GMP_RNDU);
}

// unit-roundoff macros...

#define get_u(U,Z)  mpfr_set_ui_2exp(U,1,mpfr_get_exp(Z->x)-mpfr_get_prec(Z->x),GMP_RNDU);
#define get_u2(U,Z) mpfr_set_ui_2exp(U,1,mpfr_get_exp(Z->x)-mpfr_get_prec(Z->x)-1,GMP_RNDU);
#define get_1pcplus(C,X) mpfr_mul_2si(C,X->d,1-mpfr_get_exp(X->x),GMP_RNDU); mpfr_add_ui(C,C,2,GMP_RNDU);

/* value & error-bound functions */

/* constants */

#define MPFRE_CONST(c)\
void mpfre_const_##c(mpfre_t z) {\
  mpfr_const_##c(z->x,GMP_RNDN);\
  get_u2(z->d,z)\
}

MPFRE_CONST(pi)
MPFRE_CONST(euler)

/* unary */

#define MPFRE_UNARY(f)\
void mpfre_##f(mpfre_t z, mpfre_t x) {\
  mpfr_##f(z->x,x->x,GMP_RNDN);\
  get_u2(z->d,z)\
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);\
}

MPFRE_UNARY(abs)
MPFRE_UNARY(neg)

void mpfre_sqr(mpfre_t z, mpfre_t x) { // FIXME is this right?
  MPFR_DECL_INIT(e,error_prec);
  mpfr_sqr(z->x,x->x,GMP_RNDN);
  // u(z)
  get_u(z->d,z)
  // error from x
  get_1pcplus(e,x)
  mpfr_mul(e,e,x->x,GMP_RNDU);
  mpfr_abs(e,e,GMP_RNDU);
  mpfr_mul(e,e,x->d,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
  mpfr_mul_2ui(z->d,z->d,1,GMP_RNDU);
  // total error
  mpfr_div_2ui(z->d,z->d,1,GMP_RNDU);
}

void mpfre_zeta(mpfre_t z, mpfre_t x) { // FIXME 
  fprintf(stderr,"Warning - using zeta - mpfre function not yet implemented\n");
  mpfr_zeta(z->x,x->x,GMP_RNDN);
  // u(z)
  get_u(z->d,z)
  // error from x
}

void mpfre_gamma(mpfre_t z, mpfre_t x) { // FIXME 
  fprintf(stderr,"Warning - using gamma - mpfre function not yet implemented\n");
  mpfr_gamma(z->x,x->x,GMP_RNDN);
  // u(z)
  get_u(z->d,z)
  // error from x
}

/* binary */

void mpfre_add(mpfre_t z, mpfre_t x, mpfre_t y) {
  mpfr_add(z->x,x->x,y->x,GMP_RNDN);
  if (mpfr_zero_p(z->x)) {
    mpfr_add(z->d,x->d,y->d,GMP_RNDU); // ??
    return;
  }
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);
  mpfr_add(z->d,z->d,y->d,GMP_RNDU);
}

void mpfre_sub(mpfre_t z, mpfre_t x, mpfre_t y) {
  mpfr_sub(z->x,x->x,y->x,GMP_RNDN);
  if (mpfr_zero_p(z->x)) {
    mpfr_add(z->d,x->d,y->d,GMP_RNDU); // ??
    return;
  }
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);
  mpfr_add(z->d,z->d,y->d,GMP_RNDU);
}

void mpfre_mul(mpfre_t z, mpfre_t x, mpfre_t y) {
  MPFR_DECL_INIT(e,error_prec);
  if (mpfr_zero_p(x->x)||mpfr_zero_p(y->x)) {
    mpfr_set_ui(z->x,0,GMP_RNDN);
    mpfr_set_ui(z->d,0,GMP_RNDN);
    return;
  }
  mpfr_mul(z->x,x->x,y->x,GMP_RNDN);
  // u(z)
  get_u(z->d,z)
  // error from y
  get_1pcplus(e,x)
  mpfr_mul(e,e,x->x,GMP_RNDU);
  mpfr_abs(e,e,GMP_RNDU);
  mpfr_mul(e,e,y->d,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
  // error from x
  get_1pcplus(e,y)
  mpfr_mul(e,e,y->x,GMP_RNDU);
  mpfr_abs(e,e,GMP_RNDU);
  mpfr_mul(e,e,x->d,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
  // total error
  mpfr_div_2ui(z->d,z->d,1,GMP_RNDU);
}

void mpfre_div(mpfre_t z, mpfre_t x, mpfre_t y) {
  MPFR_DECL_INIT(e,error_prec);
  if (mpfr_zero_p(x->x)) {
    mpfr_set_ui(z->x,0,GMP_RNDN);
    mpfr_set_ui(z->d,0,GMP_RNDN);
    return;
  }
  mpfr_div(z->x,x->x,y->x,GMP_RNDN);
  // u(z)
  get_u(z->d,z)
  // error from x and y
  mpfr_mul(e,z->x,y->d,GMP_RNDU);
  mpfr_abs(e,e,GMP_RNDU);
  mpfr_mul_ui(e,e,4,GMP_RNDU);
  mpfr_add(e,e,x->d,GMP_RNDU);
  mpfr_div(e,e,y->x,GMP_RNDU);
  mpfr_abs(e,e,GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

/* mpfre_add_const */

#define MPFRE_ADD_CONST(TYPE,CTYPE)\
void mpfre_add_##TYPE(mpfre_t z, mpfre_t x, CTYPE y) {\
  mpfr_add_##TYPE(z->x,x->x,y,GMP_RNDN);\
  get_u2(z->d,z)\
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);\
}

MPFRE_ADD_CONST(ui,unsigned int)
MPFRE_ADD_CONST(si,int)

/* mpfre_const_add */

#define MPFRE_CONST_ADD(TYPE,CTYPE)\
void mpfre_##TYPE##_add(mpfre_t z, CTYPE x, mpfre_t y) {\
  mpfr_add_##TYPE(z->x,y->x,x,GMP_RNDN); /* flipped args */\
  get_u2(z->d,z)\
  mpfr_add(z->d,z->d,y->d,GMP_RNDU);\
}

MPFRE_CONST_ADD(ui,unsigned int)
MPFRE_CONST_ADD(si,int)

/* mpfre_sub_const */

#define MPFRE_SUB_CONST(TYPE,CTYPE)\
void mpfre_sub_##TYPE(mpfre_t z, mpfre_t x, CTYPE y) {\
  mpfr_sub_##TYPE(z->x,x->x,y,GMP_RNDN);\
  get_u2(z->d,z)\
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);\
}

MPFRE_SUB_CONST(ui,unsigned int)
MPFRE_SUB_CONST(si,int)

/* mpfre_const_sub */

#define MPFRE_CONST_SUB(TYPE,CTYPE)\
void mpfre_##TYPE##_sub(mpfre_t z, CTYPE x, mpfre_t y) {\
  mpfr_##TYPE##_sub(z->x,x,y->x,GMP_RNDN);\
  get_u2(z->d,z)\
  mpfr_add(z->d,z->d,y->d,GMP_RNDU);\
}

MPFRE_CONST_SUB(ui,unsigned int)
MPFRE_CONST_SUB(si,int)

/* mpfre_mul_const */

void mpfre_mul_ui(mpfre_t z, const mpfre_t x, unsigned int y) {
  // z=x ok
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_mul_ui(z->x,x->x,y,GMP_RNDN);
  get_u2(t,z) // u(z)/2
  mpfr_mul_ui(e,x->d,y,GMP_RNDU); // error from x
  mpfr_add(z->d,t,e,GMP_RNDU); // total error
}

void mpfre_mul_si(mpfre_t z, const mpfre_t x, int y) {
  // z=x ok
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_mul_si(z->x,x->x,y,GMP_RNDN);
  get_u2(t,z) // u(z)/2
  mpfr_mul_ui(e,x->d,abs(y),GMP_RNDU); // error from x
  mpfr_add(z->d,t,e,GMP_RNDU); // total error
}

void mpfre_mul_qui(mpfre_t z, const mpfre_t x, unsigned int n, unsigned int d) {
  // z=x*n/d
  // z=x ok
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  //printf("A n=%u d=%u z=%p z->x=%p x->x=%p\n",n,d,z,z->x,x->x);
  mpfr_mul_ui(z->x,x->x,n,GMP_RNDN);
  //printf("B n=%u d=%u\n",n,d);
  mpfr_div_ui(z->x,z->x,d,GMP_RNDN);
  get_u2(t,z) // u(z)/2
  mpfr_mul_ui(e,x->d,n,GMP_RNDU); // error from x
  mpfr_div_ui(e,e,d,GMP_RNDU); // error from x
  mpfr_add(z->d,t,e,GMP_RNDU); // total error
  //printf("B n=%u d=%u\n",n,d);
}

/* mpfre_const_mul */

void mpfre_ui_mul(mpfre_t z, unsigned int x, const mpfre_t y) {
  mpfre_mul_ui(z,y,x);
}

void mpfre_si_mul(mpfre_t z, int x, const mpfre_t y) {
  mpfre_mul_si(z,y,x);
}

/* mpfre_div_const */

void mpfre_div_ui(mpfre_t z, mpfre_t x, unsigned int y) {
  MPFR_DECL_INIT(e,error_prec);
  mpfr_div_ui(z->x,x->x,y,GMP_RNDN);
  // u(z)/2
  get_u2(z->d,z)
  // error from x
  mpfr_div_ui(e,x->d,y,GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_div_si(mpfre_t z, mpfre_t x, int y) {
  MPFR_DECL_INIT(e,error_prec);
  mpfr_div_si(z->x,x->x,y,GMP_RNDN);
  // u(z)/2
  get_u2(z->d,z)
  // error from x
  mpfr_div_ui(e,x->d,abs(y),GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

/* mpfre_const_div */

void mpfre_ui_div(mpfre_t z, unsigned int y, mpfre_t x) {
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_ui_div(z->x,y,x->x,GMP_RNDN);
  // u(z)
  get_u2(z->d,z)
  // error from x
  mpfr_mul_ui(e,x->d,2*y,GMP_RNDU);
  mpfr_sqr(t,x->x,GMP_RNDD);
  mpfr_div(e,e,t,GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_si_div(mpfre_t z, int y, mpfre_t x) {
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_si_div(z->x,y,x->x,GMP_RNDN);
  // u(z)
  get_u2(z->d,z)
  // error from x
  mpfr_mul_ui(e,x->d,2*abs(y),GMP_RNDU);
  mpfr_sqr(t,x->x,GMP_RNDD);
  mpfr_div(e,e,t,GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_recip(mpfre_t z, mpfre_t x) {
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_ui_div(z->x,1,x->x,GMP_RNDN);
  // u(z)
  get_u2(z->d,z)
  // error from x
  mpfr_mul_ui(e,x->d,2,GMP_RNDU);
  mpfr_sqr(t,x->x,GMP_RNDD);
  mpfr_div(e,e,t,GMP_RNDU);
  // total error
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_sqrt(mpfre_t z, mpfre_t x) {
  // $z=\N(x^{1/2})$: $\delta_z \leqslant u_z/2 + \delta_x/x^{1/2} $ 
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  mpfr_sqrt(z->x,x->x,GMP_RNDN);
  mpfr_sqrt(e,x->x,GMP_RNDD);
  if (1) { // optional: more work, smaller error bound (by a factor about 2)
    //better bound: $\delta_z\leq u_z/2+\delta_x/x^{1/2}/(1+sqrt(1-2^{1-p_x}))$ 
    mpfr_set_ui_2exp(t,1,1-mpfr_get_prec(x->x),GMP_RNDU);
    mpfr_ui_sub(t,1,t,GMP_RNDD);
    mpfr_sqrt(t,t,GMP_RNDD);
    mpfr_add_ui(t,t,1,GMP_RNDD);
    mpfr_div(e,e,t,GMP_RNDU);
  }
  mpfr_div(e,x->d,e,GMP_RNDU);
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_root(mpfre_t z, mpfre_t x, unsigned int k) {
  // FIXME not checked
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  //printf("z=%p k=%u prec=%d x=",z,k,(int)mpfr_get_prec(x->x))+mpfr_out_str(stdout,10,10,x->x,GMP_RNDN)+printf("\n");
  if (k<10)
    mpfr_root(z->x,x->x,(unsigned long)k,GMP_RNDN);
  else {
    mpfr_log(z->x,x->x,GMP_RNDN);
    mpfr_div_ui(z->x,z->x,k,GMP_RNDN);
    mpfr_exp(z->x,z->x,GMP_RNDN);
  }
  //printf("z=")+mpfr_out_str(stdout,10,10,z->x,GMP_RNDN)+printf("\n");
  get_u2(z->d,z)
  mpfr_add(t,x->x,x->d,GMP_RNDU);
  mpfr_root(e,t,k,GMP_RNDU);
  mpfr_div(e,e,t,GMP_RNDU);
  mpfr_mul(e,e,x->d,GMP_RNDU);
  mpfr_div_ui(e,e,k,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_log(mpfre_t z, mpfre_t x) {
  MPFR_DECL_INIT(e,error_prec);
  mpfr_log(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z) //FIXME x->x=1 => z->x=0 case
  mpfr_div(e,x->d,x->x,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_exp(mpfre_t z, mpfre_t x) {
  MPFR_DECL_INIT(e,error_prec);
  mpfr_exp(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z)
  mpfr_add(e,x->d,x->x,GMP_RNDU);
  mpfr_exp(e,e,GMP_RNDU);
  mpfr_mul(e,x->d,e,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

/* trig */

void mpfre_sin(mpfre_t z, mpfre_t x) {
  mpfr_sin(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);
}

void mpfre_cos(mpfre_t z, mpfre_t x) {
  mpfr_cos(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,x->d,GMP_RNDU);
}

void mpfre_tan(mpfre_t z, mpfre_t x) {
  mpfre_t s=mpfre_init(mpfr_get_prec(x->x));
  mpfre_t c=mpfre_init(mpfr_get_prec(x->x));
  mpfre_sin(s,x);
  mpfre_sin(c,x);
  mpfre_div(z,s,c);
  mpfre_clear(s);
  mpfre_clear(c);
}

void mpfre_atan(mpfre_t z, mpfre_t x) { // FIXME is this better?
  MPFR_DECL_INIT(e,error_prec);
  mpfr_atan(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z)
  mpfr_sub(e,x->x,x->d,GMP_RNDD);
  mpfr_sqr(e,e,GMP_RNDD);
  mpfr_add_ui(e,e,1,GMP_RNDD);
  mpfr_div(e,x->d,e,GMP_RNDU);
  mpfr_add(z->d,z->d,e,GMP_RNDU);
}

void mpfre_atan1(mpfre_t z, mpfre_t x) { // FIXME is this better?
  mpfr_atan(z->x,x->x,GMP_RNDN);
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,x->d,GMP_RNDU); // slope<=1 for all x
}

void mpfre_asin(mpfre_t z, mpfre_t x) { // Ughhh
  mpfre_t t1=mpfre_init(mpfr_get_prec(x->x));
  mpfre_t t2=mpfre_init(mpfr_get_prec(x->x));
  mpfr_asin(z->x,x->x,GMP_RNDN);
  mpfre_mul(t1,x,x);
  mpfre_ui_sub(t2,1,t1);
  mpfre_sqrt(t1,t2);
  mpfre_div(t2,x,t1);
  mpfre_atan(z,t2);
  mpfre_clear(t1);
  mpfre_clear(t2);
}

void mpfre_acos(mpfre_t z, mpfre_t x) { // Ughhh
  mpfre_t t1=mpfre_init(mpfr_get_prec(x->x));
  mpfre_t t2=mpfre_init(mpfr_get_prec(x->x));
  mpfr_acos(z->x,x->x,GMP_RNDN);
  mpfre_mul(t1,x,x);
  mpfre_ui_sub(t2,1,t1);
  mpfre_sqrt(t1,t2);
  mpfre_div(t2,t1,x);
  mpfre_atan(z,t2);
  mpfre_clear(t1);
  mpfre_clear(t2);
}

void mpfre_erf(mpfre_t z, mpfre_t x) {
  mpfr_t e,t;
  mpfr_init2(e,error_prec);
  mpfr_init2(t,error_prec);
  mpfr_erf(z->x,x->x,GMP_RNDN);
  mpfr_sqr(e,x->x,GMP_RNDD);
  mpfr_neg(e,e,GMP_RNDD);
  mpfr_exp(e,e,GMP_RNDU); // exp(-x^2)
  mpfr_mul_2ui(e,e,1,GMP_RNDU);
  mpfr_const_pi(t,GMP_RNDD);
  mpfr_sqrt(t,t,GMP_RNDD); // sqrt(pi)
  mpfr_div(e,e,t,GMP_RNDU); // 2*exp(-x^2)/sqrt(pi)
  mpfr_mul(e,e,x->d,GMP_RNDU); // 2*exp(-x^2)/sqrt(pi)*delta(x)
  get_u2(z->d,z)
  mpfr_add(z->d,z->d,e,GMP_RNDU);
  mpfr_clear(e);
  mpfr_clear(t);
}

/* sum */

void mpfre_sum(mpfre_t s, mpfre_t x[], unsigned int n) {
  // N. Higham "Accuracy and stability of numerical algorithms" exercise 4.3
  // returns s s.t. |s->x-sum(x[i],i=0..n-1)|<s->d
  // s and d are computed to their initialized precision
  int i;
  mpfr_t u,g,t;
  mpfr_set_ui(s->x,0,GMP_RNDN);
  mpfr_set_ui(s->d,0,GMP_RNDU);
  if (n==0) return; 
  for (i=0; i<n; i++) {
    mpfr_add(s->x,s->x,x[i]->x,GMP_RNDN);
    mpfr_add(s->d,s->d,x[i]->d,GMP_RNDU);
  }
  mpfr_init2(u,error_prec);
  mpfr_init2(g,error_prec);
  mpfr_init2(t,error_prec);
  get_u(u,s) // unit roundoff
  mpfr_mul_ui(t,u,n-2,GMP_RNDU);
  mpfr_set(g,t,GMP_RNDU);
  mpfr_ui_sub(t,1,t,GMP_RNDD);
  mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-2}
  mpfr_abs(t,x[0]->x,GMP_RNDU);
  mpfr_mul(t,t,g,GMP_RNDU);
  mpfr_add(s->d,s->d,t,GMP_RNDU);
  if (n>1) {
    mpfr_abs(t,x[1]->x,GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(s->d,s->d,t,GMP_RNDU);
  }
  for (i=2; i<n; i++) {
    mpfr_mul_ui(t,u,n-i,GMP_RNDU);
    mpfr_set(g,t,GMP_RNDU);
    mpfr_ui_sub(t,1,t,GMP_RNDD);
    mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-i}
    mpfr_abs(t,x[i]->x,GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(s->d,s->d,t,GMP_RNDU);
  }
  mpfr_clear(u);
  mpfr_clear(g);
  mpfr_clear(t);
}

void mpfre_logistic(mpfre_t z, const mpfre_t x, unsigned int n, unsigned int d) {
  // z=x*(1-x)*n/d, assuming 0<=x<=1, n/d<4, n>=0, d>0
  MPFR_DECL_INIT(e,error_prec);
  MPFR_DECL_INIT(t,error_prec);
  //MPFR_DECL_INIT(t,mpfr_get_prec(x->x));
  mpfr_ui_sub(z->x,1,x->x,GMP_RNDN);
  mpfr_mul(z->x,z->x,x->x,GMP_RNDN);
  mpfr_mul_ui(z->x,z->x,n,GMP_RNDN);
  mpfr_div_ui(z->x,z->x,d,GMP_RNDN);
  get_u2(t,z) // u(z)/2 FIXME wrong - 4 roundoffs
  mpfr_mul_ui(e,x->d,n,GMP_RNDU); // error from x
  mpfr_div_ui(e,e,d,GMP_RNDU); // error from x
  mpfr_add(z->d,t,e,GMP_RNDU); // total error
}

#undef get_u
#undef get_u2
#undef get_1pcplus

#ifdef TEST_MPFRE

void sum() {
  //evalf[200](add(1/i,i=1..1000));
  //7.4854708605503449126565182043339001765216791697088036657736267499576993491652024409599344374118450813967980143822544037158148421958847034043140398433689296639178338273590557913000715469268403259337980
  int i,n=1000;
  mpfre_t s,x[n],y[n];
  mpfr_t  ex,diff;
  mpfr_init2(ex,200*3);
  mpfr_init2(diff,100);
  mpfr_set_str(ex,"7.4854708605503449126565182043339001765216791697088036657736267499576993491652024409599344374118450813967980143822544037158148421958847034043140398433689296639178338273590557913000715469268403259337980",10,GMP_RNDN);
  s=mpfre_init(100);
  for (i=0; i<n; i++) x[i]=mpfre_init(100);
  for (i=0; i<n; i++) y[i]=mpfre_init(100);
  for (i=0; i<n; i++) mpfre_set_ui(x[i],i+1,0.0);
  for (i=0; i<n; i++) mpfre_ui_div(y[n-1-i],1,x[i]);
  mpfre_sum(s,y,n);
  printf("sum=")+mpfre_out_str(s)+printf("\n");
  mpfr_sub(diff,ex,s->x,GMP_RNDN);
  printf("approx error=")+mpfr_out_str(stdout,10,10,diff,GMP_RNDN)+printf("\n");
}

void logistic() {
  int i;
  mpfre_t x0,x1,t1;
  mpfr_t y0,y1,d;
  x0=mpfre_init(300);
  x1=mpfre_init(300);
  t1=mpfre_init(300);
  mpfr_init2(y0,2000);
  mpfr_init2(y1,2000);
  mpfr_init2(d,100);
  mpfre_set_ui(x0,1,0.0);
  mpfre_div_ui(x0,x0,2);
  mpfr_set_d(y0,0.5,GMP_RNDN);
  for (i=1; i<=100; i++) {
    mpfre_ui_sub(x1,1,x0);
    mpfre_mul(t1,x1,x0);
    //mpfre_mul_ui(x1,t1,15); mpfre_div_ui(x0,x1,4);
    mpfre_mul_qui(x0,t1,15,4);
    mpfr_ui_sub(y1,1,y0,GMP_RNDN);
    mpfr_mul(y1,y1,y0,GMP_RNDN);
    mpfr_mul_ui(y1,y1,15,GMP_RNDN);
    mpfr_div_ui(y0,y1,4,GMP_RNDN);
    mpfr_sub(d,y0,x0->x,GMP_RNDN); // actual error
    printf("x%d            =",i)+mpfre_out_str(x0)+printf(" d=")+mpfr_out_str(stdout,10,8,d,GMP_RNDU)+printf("\n");
  }
}

int main() {
  mpfre_t x,y,z;
  x=mpfre_init(130);
  y=mpfre_init(200);
  z=mpfre_init(100);
  mpfre_set_si(x,1,0.0e-20);
  printf("x=")+mpfre_out_str(x)+printf("\n");
  mpfre_set_ui(y,3,1.0e-20);
  mpfre_add(z,x,y); printf("z=x+y=")+mpfre_out_str(z)+printf("\n");
  mpfre_add_ui(z,x,1000); printf("z=x+1000=")+mpfre_out_str(z)+printf("\n");
  mpfre_mul(z,x,y); printf("z=x*y=")+mpfre_out_str(z)+printf("\n");
  mpfre_div(z,x,y); printf("z=x/y=")+mpfre_out_str(z)+printf("\n");
  mpfre_mul_ui(z,x,10); printf("z=x*10=")+mpfre_out_str(z)+printf("\n");
  mpfre_div_ui(z,x,3); printf("z=x/3=")+mpfre_out_str(z)+printf("\n");
  mpfre_sqrt(z,y); printf("z=sqrt(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_log(z,y); printf("z=log(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_recip(z,y); printf("z=recip(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_ui_div(z,1000,y); printf("z=1000/y=")+mpfre_out_str(z)+printf("\n");
  mpfre_exp(z,y); printf("z=exp(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_sin(z,y); printf("z=sin(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_atan(z,y); printf("z=atan(y)=")+mpfre_out_str(z)+printf("\n");
  mpfre_const_pi(z); printf("z=pi=")+mpfre_out_str(z)+printf("\n");
  mpfre_const_euler(z); printf("z=euler=")+mpfre_out_str(z)+printf("\n");
  sum();
  logistic();
  return 0;
  mpfre_sub(z,x,x); mpfre_out_str(z)+printf("\n");
  mpfre_mul_ui(z,x,1000); mpfre_out_str(z)+printf("\n");
  mpfre_recip(z,x); mpfre_out_str(z)+printf("\n");
  mpfre_recip(z,y); mpfre_out_str(z)+printf("\n");
  mpfre_sqrt(z,y); mpfre_out_str(z)+printf("\n");
  return 0;
}
#endif
