/* K M Briggs 2006 Jun 27
 mpfs exact arithmetic 
 TODO: 
  lightweight abs,neg
  pow_ui
  recip
 FIXME
  get rid of mpfr_t->q
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <sys/times.h>

#include "mpfs.h"

int mpfs_dump(char* pre, mpfs_t x, char* post, unsigned int dig);

static double alpha=0.2;

static void eval(mpfs_t, mp_prec_t);

void mpfs_set_alpha(const double a) {
  alpha=a;
}

double mpfs_get_alpha(void) {
  return alpha;
}

const char* mpfs_get_version(void) {
  const char* mpfs_version="mpfs-0.9 compiled" " "__DATE__" "__TIME__;
  return mpfs_version;
}

void mpfre_sum_cache(mpfre_t s, mpfs_t x[], unsigned int n) {
  // special version of mpfre_sum from mpfre.c
  // N. Higham "Accuracy and stability of numerical algorithms" exercise 4.3
  // returns s s.t. |s->x-sum(x[i],i=0..n-1)|<s->d
  // s and d are computed to their initialized precision
  static const mp_prec_t error_prec=32;
  int i;
  mpfr_t u,g,t;
  mpfr_set_ui(s->x,0,GMP_RNDN);
  mpfr_set_ui(s->d,0,GMP_RNDU);
  if (n==0) return; 
  for (i=0; i<n; i++) {
    mpfr_add(s->x,s->x,x[i]->cache->x,GMP_RNDN);
    mpfr_add(s->d,s->d,x[i]->cache->d,GMP_RNDU);
  }
  mpfr_init2(u,error_prec);
  mpfr_init2(g,error_prec);
  mpfr_init2(t,error_prec);
  mpfr_set_ui_2exp(u,1,mpfr_get_exp(s->x)-mpfr_get_prec(s->x),GMP_RNDU); // unit roundoff
  mpfr_mul_ui(t,u,n-2,GMP_RNDU);
  mpfr_set(g,t,GMP_RNDU);
  mpfr_ui_sub(t,1,t,GMP_RNDD);
  mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-2}
  mpfr_abs(t,x[0]->cache->x,GMP_RNDU);
  mpfr_mul(t,t,g,GMP_RNDU);
  mpfr_add(s->d,s->d,t,GMP_RNDU);
  if (n>1) {
    mpfr_abs(t,x[1]->cache->x,GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(s->d,s->d,t,GMP_RNDU);
  }
  for (i=2; i<n; i++) {
    mpfr_mul_ui(t,u,n-i,GMP_RNDU);
    mpfr_set(g,t,GMP_RNDU);
    mpfr_ui_sub(t,1,t,GMP_RNDD);
    mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-i}
    mpfr_abs(t,x[i]->cache->x,GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(s->d,s->d,t,GMP_RNDU);
  }
  mpfr_clear(u);
  mpfr_clear(g);
  mpfr_clear(t);
}

#define MPFR_FUN_D(fun)\
static int mpfr_##fun##_d(mpfr_t y, mpfr_t x, double d, int rnd_mode) {\
  MPFR_DECL_INIT(t,53);\
  mpfr_set_d(t,d,GMP_RNDZ); /* exact */\
  return mpfr_##fun(y,x,t,rnd_mode);\
}

#define MPFR_D_FUN(fun)\
static int mpfr_d_##fun(mpfr_t y, double d, mpfr_t x, int rnd_mode) {\
  MPFR_DECL_INIT(t,53);\
  mpfr_set_d(t,d,GMP_RNDZ); /* exact */\
  return mpfr_##fun(y,t,x,rnd_mode);\
}

//MPFR_FUN_D(add)
//MPFR_FUN_D(sub)
//MPFR_FUN_D(mul)
MPFR_FUN_D(div)

//MPFR_D_FUN(add)
//MPFR_D_FUN(sub)
//MPFR_D_FUN(mul)
//MPFR_D_FUN(div)

/* macros to make commutative mpfr functions with int argument first */

#define MPFR_COMM_SI_FUN(fun)\
static int mpfr_si_##fun(mpfr_t y, long l, mpfr_t x, int rnd_mode) {\
  return mpfr_##fun##_si(y,x,l,rnd_mode);\
}

#define MPFR_COMM_UI_FUN(fun)\
static int mpfr_ui_##fun(mpfr_t y, unsigned long l, mpfr_t x, int rnd_mode) {\
  return mpfr_##fun##_ui(y,x,l,rnd_mode);\
}

//MPFR_COMM_SI_FUN(add)
//MPFR_COMM_SI_FUN(mul)

/*
  macros for cases in the switch in eval
*/

#define ONARY_CASE(X,Y)\
    case X: {\
      mpfre##Y(z->cache);\
      break;\
    }

#define UNARY_CASE(X)\
    case X: {\
      eval(z->x,prec+prec_inc);\
      mpfre##X(z->cache,z->x->cache);\
      break;\
    }

#define BINARY_CASE(X)\
    case X: {\
      eval(z->x,prec+prec_inc);\
      eval(z->y,prec+prec_inc);\
      mpfre##X(z->cache,z->x->cache,z->y->cache);\
      break;\
    }

// e.g. CONST_BINARY_CASE(si,int,add) => mpfre_add_si(z,q,y)
#define CONST_BINARY_CASE(TYPE,CTYPE,OP)\
    case _##TYPE##_##OP: {\
      eval(z->y,prec+prec_inc);\
      mpfre_##TYPE##_##OP(z->cache,*(CTYPE*)(z->q),z->y->cache);\
      break;\
    }

// e.g. BINARY_CONST_CASE(si,int,add) => mpfre_si_add(z,x,q)
#define BINARY_CONST_CASE(TYPE,CTYPE,OP)\
    case _##OP##_##TYPE: {\
      eval(z->x,prec+prec_inc);\
      mpfre_##OP##_##TYPE(z->cache,z->x->cache,*(CTYPE*)(z->q));\
      break;\
    }


/*! this function recursively evaluates z to prec bits */
static void eval(mpfs_t z, const mp_prec_t prec) {
  mp_prec_t prec_inc=0;
  if (z->f==_d || z->f==_si || z->f==_ui) return;
  if (z->cache==NULL) { /* never been evaluated */
    z->cache=malloc(sizeof(struct _mpfre));\
    mpfre_init2(z->cache,prec);
  } else {
    if (mpfr_get_prec(z->cache->x)>=prec) return;
  }
  //mpfr_init2(e,error_prec);
  //mpfr_init2(u,error_prec);
  if (alpha>=1.0)
    prec_inc=(int)alpha;
  else if (rand()<alpha*(double)RAND_MAX)
    prec_inc=1;
  mpfr_set_prec(z->cache->x,prec+prec_inc);
  switch (z->f) {
    case _si: case _ui: case _d: break; /* never get here */
    ONARY_CASE(_const_pi,_const_pi)
    UNARY_CASE(_abs)
    UNARY_CASE(_neg)
    UNARY_CASE(_sqrt)
    UNARY_CASE(_exp)
    //UNARY_CASE(_expm1,mpfr_exp)
    UNARY_CASE(_log)
    //UNARY_CASE(_log1p,mpfr_log)
    UNARY_CASE(_sin)
    UNARY_CASE(_cos)
    UNARY_CASE(_tan)
    UNARY_CASE(_asin)
    UNARY_CASE(_acos)
    UNARY_CASE(_atan)
    UNARY_CASE(_erf)
    UNARY_CASE(_gamma)
    //UNARY_CASE(_lngamma,mpfr_lngamma)
    UNARY_CASE(_zeta)
    case _set_qui: { // FIXME is this right?
      mpfr_set_ui(z->cache->x,*(unsigned int*)z->x,GMP_RNDN);
      mpfr_div_ui(z->cache->x,z->cache->x,*(unsigned int*)z->y,GMP_RNDN);
      mpfr_set_ui_2exp(z->cache->d,1,mpfr_get_exp(z->cache->x)-mpfr_get_prec(z->cache->x),GMP_RNDU);
      break;
    }
    case _mul_qui: {
      eval(z->x,prec+prec_inc);
      mpfre_mul_qui(z->cache,z->x->cache,((unsigned int*)z->y)[0],((unsigned int*)z->y)[1]);
      break;
    }
    case _root: {
      eval(z->x,prec+prec_inc);
      mpfre_root(z->cache,z->x->cache,*(unsigned int*)z->y);
      break;
    }
    BINARY_CASE(_add)
    BINARY_CASE(_sub)
    BINARY_CASE(_mul)
    BINARY_CASE(_div)
    CONST_BINARY_CASE(ui,unsigned int,sub)
    CONST_BINARY_CASE(ui,unsigned int,div)
    CONST_BINARY_CASE(si,int,sub)
    CONST_BINARY_CASE(si,int,div)
    BINARY_CONST_CASE(ui,unsigned int,add)
    BINARY_CONST_CASE(ui,unsigned int,sub)
    BINARY_CONST_CASE(ui,unsigned int,mul)
    BINARY_CONST_CASE(ui,unsigned int,div)
    BINARY_CONST_CASE(si,int,add)
    BINARY_CONST_CASE(si,int,sub)
    BINARY_CONST_CASE(si,int,mul)
    BINARY_CONST_CASE(si,int,div)
    //CONST_BINARY_CASE(d,double,div)
    //BINARY_CONSTSI_CASE(_mul_si,mpfr_mul_si)
    //BINARY_CONSTSI_CASE(_div_si,mpfr_div_si)
    //BINARY_CONSTUI_CASE(_add_ui,mpfr_add_ui)
    //BINARY_CONSTUI_CASE(_mul_ui,mpfr_mul_ui)
    //BINARY_CONSTUI_CASE(_div_ui,mpfr_div_ui)
    //CONSTUI_BINARY_CASE(_ui_div,mpfr_ui_div)
    //CONSTUI_BINARY_CASE(_ui_sub,mpfr_ui_sub)
    //CONSTD_BINARY_CASE(_d_add,mpfr_d_add)
    //CONSTD_BINARY_CASE(_d_sub,mpfr_d_sub)
    //CONSTD_BINARY_CASE(_d_mul,mpfr_d_mul)
    //BINARY_CONSTD_CASE(_add_d,mpfr_add_d)
    //BINARY_CONSTD_CASE(_sub_d,mpfr_sub_d)
    //BINARY_CONSTD_CASE(_mul_d,mpfr_mul_d)
    //BINARY_CONSTD_CASE(_div_d,mpfr_div_d)
    //case _pow_si:
    //  printf("mpfs: pow_si not yet implemented, sorry.\n"); exit(3);
    //  break;
    case _sum: {
      // remember z->x=(mpfs_t)(void*)x;
      int i,n=*(unsigned int*)(z->q);
      for (i=0; i<n; i++) eval(((mpfs_t*)z->x)[i],prec+prec_inc);
      mpfre_sum_cache(z->cache,(mpfs_t*)z->x,n);
      break;
    }
    case _logistic: {
      eval(z->x,prec+prec_inc);
      mpfre_logistic(z->cache,z->x->cache,((unsigned int*)z->y)[0],((unsigned int*)z->y)[1]);
      break;
    }
    default: {
      printf("mpfs: operation not yet implemented, sorry.\n"); exit(3);
    }
  }
  if (mpfr_nan_p(z->cache->x)) {
    printf("eval: z=")+mpfr_out_str(stdout,10,10,z->cache->x,GMP_RNDN);
    printf(" prec=%ld\n",prec);
    exit(1);
  }
  if (mpfr_inf_p(z->cache->x)) {
    printf("eval: z=")+mpfr_out_str(stdout,10,10,z->cache->x,GMP_RNDN);
    printf(" prec=%ld\n",prec);
    exit(2);
  }
  //mpfr_clear(e);
  //mpfr_clear(u);
}

#define MPFS_SET_CONST(C,T)\
mpfs_t mpfs_set_##C(const T C) {\
  mpfs_t x=malloc(sizeof(struct node));\
  x->cache=malloc(sizeof(struct _mpfre));\
  mpfre_init2(x->cache,8*sizeof(T));\
  mpfr_set_##C(x->cache->x,C,GMP_RNDN);\
  mpfr_set_ui(x->cache->d,0,GMP_RNDN);\
  x->f=_##C;\
  return x;\
}

MPFS_SET_CONST(si,long)
MPFS_SET_CONST(ui,unsigned long)
MPFS_SET_CONST(d,double)

mpfs_t mpfs_set_qui(const unsigned long num, const unsigned long den) {
  mpfs_t x=malloc(sizeof(struct node));
  x->cache=NULL;
  //mpfr_init2(x->cache,16*sizeof(long));
  x->x=malloc(sizeof(unsigned long));
  x->y=malloc(sizeof(unsigned long));
  *(unsigned long*)x->x=num;
  *(unsigned long*)x->y=den;
  //mpfr_init2(x->cache->d,error_prec);
  //mpfr_set_ui(x->cache->d,0,GMP_RNDN);
  //x->inexact=1;
  x->f=_set_qui;
  //x->nref=0;
  return x;
}

#define ONARY_FUNCTION(X)\
mpfs_t mpfs_##X(void) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->f=_##X;\
  return z;\
}

#define UNARY_FUNCTION(X)\
mpfs_t mpfs_##X(const mpfs_t x) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->x=x;\
  z->f=_##X;\
  return z;\
}

#define BINARY_FUNCTION(X)\
mpfs_t mpfs_##X(const mpfs_t x, const mpfs_t y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->x=x;\
  z->y=y;\
  z->f=_##X;\
  return z;\
}

/* long x op mpfs_t y */
  /*mpfr_init_set_ui(z->cache->d,0,GMP_RNDN);*/\
#define SI_BINARY_FUNCTION(X)\
mpfs_t mpfs_##X(const long x, const mpfs_t y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(long));\
  *(long*)(z->q)=x;\
  z->y=y;\
  z->f=_##X;\
  return z;\
}

/* mpfs_t x op long y */
#define BINARY_SI_FUNCTION(X)\
mpfs_t mpfs_##X(const mpfs_t x, const long y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(long));\
  *(long*)(z->q)=y;\
  z->x=x;\
  z->f=_##X;\
  return z;\
}

/* mpfs_t x op unsigned long y */
#define BINARY_UI_FUNCTION(X)\
mpfs_t mpfs_##X(const mpfs_t x, const unsigned long y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(unsigned long));\
  *(unsigned long*)(z->q)=y;\
  z->x=x;\
  z->f=_##X;\
  return z;\
}

/* unsigned long x op mpfs_t y */
#define UI_BINARY_FUNCTION(X)\
mpfs_t mpfs_##X(const unsigned long x, const mpfs_t y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(unsigned long));\
  *(unsigned long*)(z->q)=x;\
  z->y=y;\
  z->f=_##X;\
  return z;\
}

/* double x op mpfs_t y */
#define D_BINARY_FUNCTION(X)\
mpfs_t mpfs_##X(const double x, const mpfs_t y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(int));\
  *(double*)(z->q)=x;\
  z->y=y;\
  z->f=_##X;\
  return z;\
}

/* mpfs_t x op double y */
#define BINARY_D_FUNCTION(X)\
mpfs_t mpfs_##X(const mpfs_t x, const double y) {\
  mpfs_t z=malloc(sizeof(struct node));\
  z->cache=NULL;\
  z->q=malloc(sizeof(double));\
  *(double*)(z->q)=y;\
  z->x=x;\
  z->f=_##X;\
  return z;\
}

/* z=x*n/d */
mpfs_t mpfs_mul_qui(const mpfs_t x, unsigned int n, unsigned int d) {
  mpfs_t z=malloc(sizeof(struct node));
  z->cache=NULL;
  z->x=x;
  z->y=malloc(2*sizeof(unsigned int));
  ((unsigned int*)(z->y))[0]=n;
  ((unsigned int*)(z->y))[1]=d;
  //printf("mpfs_mul_qui: n=%u d=%u z->x=%p\n",n,d,z->x);
  z->f=_mul_qui;
  return z;
}

/* z=x^(1/k) */
mpfs_t mpfs_root(const mpfs_t x, unsigned int k) {
  mpfs_t z=malloc(sizeof(struct node));
  z->cache=NULL;
  z->x=x;
  z->y=malloc(sizeof(unsigned int));
  *(unsigned int*)(z->y)=k;
  z->f=_root;
  return z;
}

/*
mpfs_t mpfs_mul_qui(const mpfs_t x, unsigned int n, unsigned int d) {
  mpfs_t z=malloc(sizeof(struct node));
  z->cache=NULL;
  z->x=x;
  z->y=malloc(sizeof(unsigned int));
  z->q=malloc(sizeof(unsigned int));
  *(unsigned int*)(z->y)=n;
  *(unsigned int*)(z->q)=d;
  printf("mpfs_mul_qui: n=%u d=%u z->x=%p z->x->cache=%p\n",n,d,z->x,z->x->cache);
  z->f=_mul_qui;
  return z;
}
*/

ONARY_FUNCTION(const_pi)

UNARY_FUNCTION(abs)
UNARY_FUNCTION(neg)
UNARY_FUNCTION(sqrt)
UNARY_FUNCTION(exp)
UNARY_FUNCTION(expm1)
UNARY_FUNCTION(log)
UNARY_FUNCTION(log1p)
UNARY_FUNCTION(sin)
UNARY_FUNCTION(cos)
UNARY_FUNCTION(tan)
UNARY_FUNCTION(asin)
UNARY_FUNCTION(acos)
UNARY_FUNCTION(atan)
UNARY_FUNCTION(erf)
UNARY_FUNCTION(gamma)
UNARY_FUNCTION(lngamma)
UNARY_FUNCTION(zeta)

BINARY_FUNCTION(add)
BINARY_FUNCTION(sub)
BINARY_FUNCTION(mul)
BINARY_FUNCTION(div)

SI_BINARY_FUNCTION(si_add)
SI_BINARY_FUNCTION(si_sub)
SI_BINARY_FUNCTION(si_mul)
SI_BINARY_FUNCTION(si_div)

BINARY_SI_FUNCTION(add_si)
BINARY_SI_FUNCTION(sub_si)
BINARY_SI_FUNCTION(mul_si)
BINARY_SI_FUNCTION(div_si)

BINARY_SI_FUNCTION(pow_si)

BINARY_UI_FUNCTION(add_ui)
BINARY_UI_FUNCTION(sub_ui)
BINARY_UI_FUNCTION(mul_ui)
BINARY_UI_FUNCTION(div_ui)

UI_BINARY_FUNCTION(ui_sub)
UI_BINARY_FUNCTION(ui_div)

BINARY_D_FUNCTION(add_d)
BINARY_D_FUNCTION(sub_d)
BINARY_D_FUNCTION(mul_d)
BINARY_D_FUNCTION(div_d)

D_BINARY_FUNCTION(d_add)
D_BINARY_FUNCTION(d_sub)
D_BINARY_FUNCTION(d_mul)
D_BINARY_FUNCTION(d_div)

/* other functions */

mpfs_t mpfs_sum(mpfs_t x[], unsigned int n) {
  mpfs_t z=malloc(sizeof(struct node));
  z->cache=NULL;
  z->q=malloc(sizeof(unsigned int));
  *(unsigned int*)(z->q)=n;
  z->x=(mpfs_t)(void*)x; // we borrow x to point to the start of the array
  z->f=_sum;
  //z->inexact=1;
  //x->nref++;
  return z;
}

mpfs_t mpfs_logistic(const mpfs_t x, unsigned int n, unsigned int d) {
  mpfs_t z=malloc(sizeof(struct node));
  z->cache=NULL;
  z->x=x;
  z->y=malloc(2*sizeof(unsigned int));
  ((unsigned int*)(z->y))[0]=n;
  ((unsigned int*)(z->y))[1]=d;
  z->f=_logistic;
  //z->inexact=1;
  //x->nref++;
  return z;
}

/* evaluation-initiating functions... */

int mpfs_cmp_d(mpfs_t x, const double y) {
  /* -1 if x<y, +1 if x>y */
  int r;
  mpfr_t bound;
  mp_prec_t prec=32;
  mpfr_inits2(prec,bound,(void*)0);
  do {
    eval(x,prec);
    mpfr_add(bound,x->cache->x,x->cache->d,GMP_RNDU); /* ub for x */
    r=mpfr_cmp_d(bound,y);
    if (r<0) break; /* x<y */
    mpfr_sub(bound,x->cache->x,x->cache->d,GMP_RNDD); /* lb for x */
    r=mpfr_cmp_d(bound,y);
    if (r>0) break; /* x>y */
    //prec+=8;
    prec*=2;
    mpfr_set_prec(bound,prec);
  } while (1);
  mpfr_clears(bound,(void*)0);
  return r;
}

int mpfs_cmp(mpfs_t x, mpfs_t y) {
  /* -1 if x<y, +1 if x>y */
  int r;
  mpfr_t dn,up;
  mp_prec_t prec=64;
  //if (!x->inexact && !y->inexact) return mpfr_cmp(x->cache,y->cache);
  mpfr_inits2(prec,dn,up,(void*)0);
  do {
    eval(x,prec);
    eval(y,prec);
    mpfr_add(up,x->cache->x,x->cache->d,GMP_RNDU); /* ub for x */
    mpfr_sub(dn,y->cache->x,y->cache->d,GMP_RNDD); /* lb for y */
    r=mpfr_cmp(up,dn);
    if (r<0) break; /* x<y */
    mpfr_sub(dn,x->cache->x,x->cache->d,GMP_RNDD); /* lb for x */
    mpfr_add(up,y->cache->x,y->cache->d,GMP_RNDU); /* ub for y */
    r=-mpfr_cmp(up,dn);
    if (r>0) break; /* x>y */
    prec+=8;
    mpfr_set_prec(dn,prec);
    mpfr_set_prec(up,prec);
  } while (1);
  mpfr_clears(dn,up,(void*)0);
  return r;
}

long mpfs_floor(mpz_t y, mpfs_t x) {
  const int dbg=0;
  mp_prec_t prec=32;
  long f;
  mpfr_t dn,up;
  mpfr_init2(dn,prec);
  mpfr_init2(up,prec);
  do {
    eval(x,prec);
    if (mpfr_cmp_ui(x->cache->d,1)<0) { /* diameter<1/2 */
      mpfr_sub(dn,x->cache->x,x->cache->d,GMP_RNDD); /* lb for x */
      mpfr_add(up,x->cache->x,x->cache->d,GMP_RNDU); /* ub for x */
      mpfr_floor(dn,dn);
      mpfr_floor(up,up);
      if (dbg) {
        printf("prec=%d x=",(int)prec)+
          mpfr_out_str(stdout,10,30,x->cache->x,GMP_RNDN)+printf(" ");
        printf("d=")+
          mpfr_out_str(stdout,10,10,x->cache->d,GMP_RNDN)+printf(" ");
        printf("dn=%.0f ",mpfr_get_d(dn,GMP_RNDN));
        printf("up=%.0f\n",mpfr_get_d(up,GMP_RNDN));
      }
      if (mpfr_cmp(dn,up)==0) {
        f=mpfr_get_si(dn,GMP_RNDD); // FIXME case when it doesn't fit in long
        mpfr_get_z(y,dn,GMP_RNDD);
        mpfr_clears(dn,up,(void*)0);
        return f;
      }
    }
    prec*=2;
    //prec+=8;
    mpfr_set_prec(dn,prec);
    mpfr_set_prec(up,prec);
  } while (1);
  return 0; /* never get here */
}

int mpfs_dump(char* pre, mpfs_t x, char* post, unsigned int dig) {
  /* dump to dig dp - meant for debugging only */
  eval(x,(mp_prec_t)ceil(3.321928094887362347870319429*dig)); /* log_2(10) */
  return
    printf("%s",pre)+
    mpfr_out_str(stdout,10,dig,x->cache->x,GMP_RNDN)+
    printf("%s",post);
}

size_t mpfs_out_str(FILE *f, int b, size_t n, mpfs_t x) {
  /* like mpfr_out_str, but no rounding-mode argument */
  double log_rel_err,logb=log(b);
  mpfr_t logx,loge;
  mpfr_init2(loge,64);
  mpfr_init2(logx,64);
  mp_prec_t prec=(mp_prec_t)ceil(3.321928094887362347870319*n); /* log2(10) */
  prec=32+4*(prec/4); // FIXME ??
  do {
    eval(x,prec);
    mpfr_abs(logx,x->cache->x,GMP_RNDD);
    mpfr_log(logx,logx,GMP_RNDD);
    mpfr_div_d(logx,logx,logb,GMP_RNDU); /* logx=log_b(|x|) */
    if (mpfr_cmp_ui(x->cache->d,0)==0) {
      log_rel_err=-1e307;
    } else {
      mpfr_log(loge,x->cache->d,GMP_RNDU);
      mpfr_div_d(loge,loge,logb,GMP_RNDU); /* loge=log_b(absolute error) */
      mpfr_sub(loge,loge,logx,GMP_RNDU);   /* loge=log_b(relative error) */
      log_rel_err=mpfr_get_d(loge,GMP_RNDU);
      prec+=32;
    }
    //fprintf(stderr,"prec=%ld, log%d_rel_err=%.6g\n",prec,b,log_rel_err);
  } while (-log_rel_err<n); /* -log_b(relative error)<digits */
  mpfr_clear(loge);
  mpfr_clear(logx);
  return
    mpfr_out_str(f,b,n,x->cache->x,GMP_RNDN)+
    fprintf(f," ± ")+
    mpfr_out_str(f,10,5,x->cache->d,GMP_RNDU);
}
