#include <stdio.h>
#include <mpfr.h>
#include "mpfre.h"

const char* mpfs_get_version(void);

typedef struct node* mpfs_t;

enum op {
  /* ONARY */
  _const_pi,
  /* UNARY */
  _si,_ui,_d,_abs,_neg,_sqr,_sqrt,_exp,_expm1,_log,_log1p,_sin,_cos,_tan,_asin,_acos,_atan,_erf,_gamma,_lngamma,_zeta,
  /* BINARY */
  _add,_sub,_mul,_div,_set_qui,
  /* SI_BINARY */
  _si_add,_si_sub,_si_mul,_si_div,
  /* BINARY_SI */
  _add_si,_sub_si,_mul_si,_div_si,
  _root,_pow_si,
  /* BINARY_UI */
  _add_ui,_sub_ui,_mul_ui,_div_ui,
  /* UI_BINARY */
  _ui_sub,_ui_div,
  /* D_BINARY */
  _d_add,_d_mul,_d_sub,_d_div,
  /* BINARY_D */
  _add_d,_sub_d,_mul_d,_div_d,
  /* TERNARY */
  _mul_qui,
  /* other */
  _sum,_logistic,
};

struct node {        /* internal representation of an mpfs_t */
  mpfs_t x;          /* pointer to left  operand */
  mpfs_t y;          /* pointer to right operand */
  void* q;           /* pointer to scalar data */
  mpfre_t cache;     /* cached value */
  //unsigned int nref; /* reference count - not used yet */
  //int inexact;       /* inexactness flag */
  enum op f;         /* operation */
};

const char* mpfs_get_version(void);
void mpfs_set_alpha(const double);
double mpfs_get_alpha(void);

/* setting */
mpfs_t mpfs_set_ui(const unsigned long);
mpfs_t mpfs_set_si(const long);
mpfs_t mpfs_set_d(const double);
mpfs_t mpfs_set_qui(const unsigned long,const unsigned long); // divides really

/* ONARY_FUNCTION */
mpfs_t mpfs_const_pi(void);

/* UNARY_FUNCTION */
mpfs_t mpfs_abs(const mpfs_t);
mpfs_t mpfs_neg(const mpfs_t);
mpfs_t mpfs_sqr(const mpfs_t);
mpfs_t mpfs_sqrt(const mpfs_t);
mpfs_t mpfs_exp(const mpfs_t);
mpfs_t mpfs_expm1(const mpfs_t);
mpfs_t mpfs_log(const mpfs_t);
mpfs_t mpfs_log1p(const mpfs_t);
mpfs_t mpfs_sin(const mpfs_t);
mpfs_t mpfs_cos(const mpfs_t);
mpfs_t mpfs_tan(const mpfs_t);
mpfs_t mpfs_asin(const mpfs_t);
mpfs_t mpfs_acos(const mpfs_t);
mpfs_t mpfs_atan(const mpfs_t);
mpfs_t mpfs_erf(const mpfs_t);
mpfs_t mpfs_gamma(const mpfs_t);
mpfs_t mpfs_lngamma(const mpfs_t);
mpfs_t mpfs_zeta(const mpfs_t);

/* BINARY_FUNCTION */
mpfs_t mpfs_add(const mpfs_t, const mpfs_t);
mpfs_t mpfs_sub(const mpfs_t, const mpfs_t);
mpfs_t mpfs_mul(const mpfs_t, const mpfs_t);
mpfs_t mpfs_div(const mpfs_t, const mpfs_t);

/* SI_BINARY_FUNCTION */
mpfs_t mpfs_si_add(const long, const mpfs_t);
mpfs_t mpfs_si_sub(const long, const mpfs_t);
mpfs_t mpfs_si_mul(const long, const mpfs_t);
mpfs_t mpfs_si_div(const long, const mpfs_t);

/* D_BINARY_FUNCTION */
mpfs_t mpfs_d_add(const double, const mpfs_t); /* not really needed */
mpfs_t mpfs_d_sub(const double, const mpfs_t);
mpfs_t mpfs_d_mul(const double, const mpfs_t); /* not really needed */
mpfs_t mpfs_d_div(const double, const mpfs_t);

/* BINARY_SI_FUNCTION */
mpfs_t mpfs_add_si(const mpfs_t, const long);
mpfs_t mpfs_sub_si(const mpfs_t, const long);
mpfs_t mpfs_mul_si(const mpfs_t, const long);
mpfs_t mpfs_div_si(const mpfs_t, const long);
mpfs_t mpfs_pow_si(const mpfs_t, const long);

/* BINARY_UI_FUNCTION */
mpfs_t mpfs_add_ui(const mpfs_t, const unsigned long);
mpfs_t mpfs_sub_ui(const mpfs_t, const unsigned long);
mpfs_t mpfs_mul_ui(const mpfs_t, const unsigned long);
mpfs_t mpfs_div_ui(const mpfs_t, const unsigned long);

/* UI_BINARY_FUNCTION */
mpfs_t mpfs_ui_sub(const unsigned long, const mpfs_t);
mpfs_t mpfs_ui_div(const unsigned long, const mpfs_t);

/* BINARY_D_FUNCTION */
mpfs_t mpfs_add_d(const mpfs_t, const double);
mpfs_t mpfs_sub_d(const mpfs_t, const double);
mpfs_t mpfs_mul_d(const mpfs_t, const double);
mpfs_t mpfs_div_d(const mpfs_t, const double);

/* other functions */
mpfs_t mpfs_root(mpfs_t x, unsigned int k);
mpfs_t mpfs_mul_qui(const mpfs_t x, unsigned int n, unsigned int d);
mpfs_t mpfs_sum(mpfs_t x[], unsigned int n);
mpfs_t mpfs_logistic(const mpfs_t x, unsigned int n, unsigned int d);

/* evaluation-initiating functions: comparison, i/o, etc. */
int mpfs_cmp_si(mpfs_t x, const long d);
int mpfs_cmp_d(mpfs_t x, const double y);
int mpfs_cmp(mpfs_t x, mpfs_t y);
long mpfs_floor(mpz_t y, mpfs_t x);
int mpfs_fprint(FILE* f, mpfs_t x, unsigned int dig);
size_t mpfs_out_str(FILE *f, int b, size_t n, mpfs_t x);
