// timing functions...

/*! \file timing.h */

#include <sys/times.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
 
//int getrusage(int who, struct rusage *usage);

static long tps; // ticks per second
static struct tms t;
static clock_t t0,t1;

/* next needed?  Same as what we already have?
static struct rusage ruse;
static double T0,T1;
#define CPUTIME (getrusage(RUSAGE_SELF,&ruse),ruse.ru_utime.tv_sec+ruse.ru_stime.tv_sec+1e-6*(ruse.ru_utime.tv_usec+ruse.ru_stime.tv_usec))
*/

//! initialize timing
double timing_init(FILE *f) {
  tps=sysconf(_SC_CLK_TCK); // ticks per second
  times(&t); 
  t0=t.tms_utime; 
  if (f) fprintf(f,"start: t0=%f seconds\n",t0/(double)tps);
  return t0/(double)tps;
}

//! return time since last call to timing_init
double timing_end(FILE *f) {
  times(&t);
  t1=t.tms_utime;
  if (f) fprintf(f,"end:   t1=%f seconds\n",t1/(double)tps);
  if (f) fprintf(f,"%.2f seconds user time elapsed\n",(double)(t1-t0)/tps);
  return (double)(t1-t0)/tps;
}

/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */

int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y) {
  /* Perform the carry for the later subtraction by updating y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
     tv_usec is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}
